/*
 * $Id: LowestAverageRTTServerSelection.java 286 2005-12-03 01:07:16Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound;

import java.util.Collections;
import java.util.List;
import java.util.Comparator;

import org.apache.log4j.Logger;

/**
 * This is a ServerSelection implementation that will choose the lowest
 * average RTT host first. It doesn't do any form of ageing or rebalancing --
 * RTTs only change in response to a Host being used.
 * 
 * @author davidb
 * @version $Revision: 286 $
 */
public class LowestAverageRTTServerSelection implements ServerSelection
{
  private Logger log = Logger.getLogger(this.getClass());

  /**
   * Standard constructor.
   */
  public LowestAverageRTTServerSelection()
  {
    log.info("using " + getClass()
        + " as the server selection implementation.");
  }

  /**
   * This inner class maintains the data necessary to calculate the average
   * RTT/
   */
  private static class HostStats
  {
    public long totalRTT;
    public int  nqueries;

    public int getAvgRTT()
    {
      if (nqueries == 0) return 0;

      return (int) (totalRTT / nqueries);
    }
  }

  /**
   * Given a host, extract the host statistics.
   * 
   * @param host The host in question.
   * @return the HostStats object (creating one if the Host didn't have one).
   */
  private static HostStats getHostStats(Host host)
  {
    HostStats stats = (HostStats) host.getState();
    if (stats == null)
    {
      stats = new HostStats();
      host.setState(stats);
    }

    return stats;
  }

  /**
   * Set the RTT for a given host.
   */
  public void setRTT(Host host, long rtt)
  {
    HostStats stats = getHostStats(host);

    // timeouts are equal to 2 seconds rtt.
    if (rtt < 0) rtt = 2000;
    stats.totalRTT += rtt;
    stats.nqueries++;
  }

  /**
   * Get the value to use for the wakeup timer.
   */
  public int getWakeupTimer(Host host)
  {
    HostStats stats = getHostStats(host);

    int rtt = stats.getAvgRTT();
    if (rtt == 0) return 100; // default for first time hosts.
    return rtt + 50;
  }

  /**
   * This is a comparator class used to order the Hosts in a list of candidate
   * hosts. This one just orders Hosts by average RTT.
   */
  private static class HostComparator implements Comparator
  {
    public int compare(Object o1, Object o2)
    {
      HostStats s1 = getHostStats((Host) o1);
      HostStats s2 = getHostStats((Host) o2);

      return s1.getAvgRTT() - s2.getAvgRTT();
    }

  }

  HostComparator hc = new HostComparator();

  /**
   * Order the list of targets into preferred order.
   */
  public void orderTargets(List targets)
  {
    Collections.sort(targets, hc);
  }

}
